#version 430
#extension GL_ARB_separate_shader_objects : enable










struct SComputeInput
{
	uvec3 GlobalThreadID; uvec3 LocalThreadID; uvec3 GroupID;};

layout( std140) uniform UCubemapData
{
	uint Face;
	float Roughness;
	uint Size;
	uint SampleCount;
	uint PerformedSampleCount;
	uint TotalSampleCount;
} CubemapData;
 uniform samplerCube InputTextureCube;
 layout(rgba16f) uniform image2D OutData;










float radicalInverse_VdC(uint bits)
{
	bits = (bits << 16u) | (bits >> 16u);
	bits = ((bits & 0x55555555u) << 1u) | ((bits & 0xAAAAAAAAu) >> 1u);
	bits = ((bits & 0x33333333u) << 2u) | ((bits & 0xCCCCCCCCu) >> 2u);
	bits = ((bits & 0x0F0F0F0Fu) << 4u) | ((bits & 0xF0F0F0F0u) >> 4u);
	bits = ((bits & 0x00FF00FFu) << 8u) | ((bits & 0xFF00FF00u) >> 8u);
	return float(bits) * 2.3283064365386963e-10;
 }


vec2 hammersley2d(uint i, uint N)
{
	return vec2(float(i) / float(N), radicalInverse_VdC(i));
}

float hash13(vec3 p3)
{
	p3 = fract(p3 * .1031);
	p3 += dot(p3, p3.yzx + 33.33);
	return fract((p3.x + p3.y) * p3.z);
}


vec3 importanceSampleGGX(vec2 Xi, vec3 N, float roughness, float rot)
{
	float a = roughness * roughness;

	float phi = 2.0 *  3.14159f  * fract(Xi.x + rot);
	float cosTheta = sqrt((1.0 - Xi.y) / (1.0 + (a*a - 1.0) * Xi.y));
	float sinTheta = sqrt(1.0 - cosTheta*cosTheta);


	vec3 H;
	H.x = cos(phi) * sinTheta;
	H.y = sin(phi) * sinTheta;
	H.z = cosTheta;


	vec3 up = abs(N.z) < 0.5 ? vec3(0.0, 0.0, 1.0) : vec3(1.0, 0.0, 0.0);
	vec3 tangent = normalize(cross(up, N));
	vec3 bitangent = cross(N, tangent);

	vec3 sampleVec = tangent * H.x + bitangent * H.y + N * H.z;

	return normalize(sampleVec);
}

float DistributionGGX(vec3 N, vec3 H, float a)
{
	float a2 = a*a;
	float NdotH = max(dot(N, H), 0.0);
	float NdotH2 = NdotH*NdotH;

	float nom = a2;
	float denom = (NdotH2 * (a2 - 1.0) + 1.0);
	denom =  3.14159f  * denom * denom;

	return nom / denom;
}

void ComputeMain( in SComputeInput  cInput  )
{
	uint face =  CubemapData.Face ;
	float roughness =  CubemapData.Roughness ;
	uint size =  CubemapData.Size ;
	uint sampleCount =  CubemapData.SampleCount ;
	uint performedSampleCount =  CubemapData.PerformedSampleCount ;
	uint totalSampleCount =  CubemapData.TotalSampleCount ;

	if (cInput.GlobalThreadID.x >= size || cInput.GlobalThreadID.y >= size) return;

	vec2 uv = vec2(1.0f, 1.0f) - (( vec2(cInput.GlobalThreadID.xy)  + vec2(0.5, 0.5)) /  float(size) );
	uv = uv * 2.f - 1.0f;

	vec3 NArr[6] =
	{
		vec3(1, uv.y, uv.x),
		vec3(-1, uv.y, -uv.x),
		vec3(-uv.x, 1, -uv.y),
		vec3(-uv.x, -1, uv.y),
		vec3(-uv.x, uv.y, 1),
		vec3(uv.x, uv.y, -1)
	};
	vec3 N = normalize(NArr[face]);
	vec3 R = N;
	vec3 V = R;

	vec3 radiance = vec3(0, 0, 0);
	float totalWeight = 0.0f;
	float rot = hash13(N * 100.0f);

	for (uint i = 0u; i < sampleCount; ++i)
	{

		vec2 Xi = hammersley2d(performedSampleCount + i, totalSampleCount);
		vec3 H = importanceSampleGGX(Xi, N, roughness, rot);
		vec3 L = normalize(2.0 * dot(V, H) * H - V);
		float NdotL = min(max(dot(N, L), 0.0f), 1.0f);

		if (NdotL > 0.0f)
		{

			float NdotH = max(dot(N, H), 0.0);

			float D = DistributionGGX(N, H, roughness);
			float HdotV = max(dot(H, V), 0.0);
			float pdf = (D * NdotH / (4.0 * HdotV)) + 0.0001;
			float resolution = 1024;
			float saTexel = 4.0 *  3.14159f  / (6.0 * resolution * resolution);
			float saSample = 1.0 / (float(totalSampleCount) * pdf + 0.0001);
			float mipLevel = max(0.5f * log2( saSample / saTexel ), 0.0f) + 1.0f;

			vec3 radianceSample = vec3(0, 0, 0);

			radianceSample =  textureLod(InputTextureCube, L, mipLevel) .rgb;

			radiance += radianceSample * NdotL;
			totalWeight += NdotL;
		}
	}

	radiance /= totalWeight;

	vec4 col = vec4(0, 0, 0, 1);
	if (performedSampleCount == 0u)
		col.rgb = radiance;
	else
	{
		vec3 prevRadiance =  imageLoad(OutData, ivec2(cInput.GlobalThreadID.xy)) .rgb;
		col.rgb = radiance *  float(sampleCount)  + prevRadiance *  float(performedSampleCount) ;
		col.rgb /=  float(sampleCount)  +  float(performedSampleCount) ;
	}

	imageStore(OutData, ivec2(cInput.GlobalThreadID.xy), col) ;
}


layout(local_size_x = 32, local_size_y = 32, local_size_z = 1) in;
void main()
{
	SComputeInput cInput;
	cInput.GlobalThreadID = gl_GlobalInvocationID;
	cInput.LocalThreadID = gl_LocalInvocationID;
	cInput.GroupID = gl_WorkGroupID;
	ComputeMain(cInput);
}
